#!/usr/bin/python

# superterm
# azz@gnu.org

import sys
import os
import select
import curses
import time
import pty
import struct
import fcntl
import IOCTLS

prefs = { "audible_beep" : 0 }
terminals = {}
activeterminal = None
quityet = 0

# Generic class to create a new Curses window
class BasicWindow:
    def __init__(self, height, width, top, left):
        self.win = curses.newwin(height, width, top, left)
        self.win.nodelay(1)
        self.win.keypad(1)
        self.win.scrollok(1)
        self.win.idlok(1)

# Clock/status window class
class StatusWindow(BasicWindow):
    def __init__(self, rootwin):
        self.root = rootwin
        rooty, rootx = self.root.getmaxyx()
        BasicWindow.__init__(self, 0, 0, rooty-1, 0)
        self.win.hline(0, 0, 0, rootx)
        self.win.addstr(0, 2, "superterm")
        self.process()
    def process(self):
        rooty, rootx = self.root.getmaxyx()
        s = time.strftime("%y/%m/%d %H:%M:%S", time.localtime(time.time()))
        self.win.addstr(0, rootx - 2 - len(s), s)
        self.win.refresh()

attributes = { "S" : curses.A_STANDOUT, \
               "U" : curses.A_UNDERLINE, \
               "R" : curses.A_REVERSE, \
               "D" : curses.A_DIM, \
               "B" : curses.A_BOLD }

# Terminal window class
class TerminalWindow(BasicWindow):
    def __init__(self, height, width, top, left, cmd):
        self.child = ChildProcess(cmd, height, width)
        BasicWindow.__init__(self, height, width, top, left)
        self.win.addstr("(Child window running " + cmd + ")")
        self.win.refresh()
        self.parsestate = -1
        self.param = ""
        self.control = ""
        self.storex = 0
        self.storey = 0
    def fileno(self):
        return self.child.masterfd
    def process(self):
        data = os.read(self.child.masterfd, 1024)
        for x in data:
            # Here is the terminal emulator.
            yy, xx = self.win.getyx()
            # This state machine handles the ^Yxx commands.
            if self.parsestate != -1:
                # State 1: one parameter to read.
                if self.parsestate == 1:
                    self.param = x
                    # Note: this triggers immediately.
                    self.parsestate = 0
                # State 5: read the commands.
                if self.parsestate == 5:
                    self.control = x
                    if x == "R" or x == "C" or x == "H" or x == "V" or x == "+" or x == "-":
                        # Fine, one parameter left to read.
                        self.parsestate = 1
                    else:
                        # What? Just ignore it.
                        self.parsestate = -1
                # State 0: actually do the command.
                if self.parsestate == 0:
                    sys.stderr.write("State 0: control=" + self.control + " param=" + str(ord(self.param)) + "\n")
                    o = ord(self.param) - 32
                    if self.control == "C" and self.param == ">":
                        self.storey, self.storey = self.win.getyx()
                    elif self.control == "C" and self.param == "<":
                        self.win.move(self.storey, self.storex)
                    elif self.control == "H":
                        if o >= 0 and o < self.win.getmaxyx()[1]:
                            self.win.move(yy, o)
                    elif self.control == "V":
                        if o >= 0 and o < self.win.getmaxyx()[0]:
                            self.win.move(o, xx)
                    elif self.control == "+":
                        if attributes.has_key(self.param):
                            self.win.attron(attributes[self.param])
                    elif self.control == "-":
                        if attributes.has_key(self.param):
                            self.win.attroff(attributes[self.param])
                        elif self.param == "A":
                            for a in attributes.keys():
                                self.win.attroff(attributes[a])
                    # Done, back to normality
                    self.parsestate = -1
                # And on to the next character.
                continue
            # cr=^M (I am certain that this is the wrong behaviour!)
            if x == chr(13):
                if yy < self.win.getmaxyx()[0] - 1:
                    self.win.move(yy, 0)
            # clear=^L
            if x == chr(12):
                self.win.clear()
            # home=^E
            elif x == chr(5):
                self.win.move(0,0)
            # ll=^F
            elif x == chr(6):
                self.win.move(self.win.getmaxyx()[0] - 1, 0)
            # bel=^G
            elif x == chr(7):
                if prefs["audible_beep"] == 1:
                    curses.beep()
                else:
                    curses.flash()
            # cub1=^H
            elif x == chr(8):
                if xx > 0:
                    self.win.move(yy, xx - 1)
            # cuf1=^N
            elif x == chr(14):
                if xx < self.win.getmaxyx()[1] - 1:
                    self.win.move(yy, xx + 1)
            # cud1=^J
            elif x == chr(10):
                if yy < self.win.getmaxyx()[0] - 1:
                    self.win.move(yy + 1, xx)
                else:
                    self.win.scroll()
                    self.win.move(self.win.getmaxyx()[0] - 1, xx)
            # cuu1=^K
            elif x == chr(11):
                if yy > 0:
                    self.win.move(yy - 1, xx)
            # extended sequences: ^Y
            elif x == chr(25):
                self.parsestate = 5
            # standard character -- pass it on through
            else:
                self.win.addch(ord(x))
        self.win.refresh()
    def feed(self, str):
        os.write(self.child.masterfd, str)

# Class for a program running in a pty -- cp.file is the file object for the pty
class ChildProcess:
    def __init__(self, cmd, height, width):
        # Get ptys
        self.masterfd, self.ptyname = pty.master_open()
        self.slavefd = pty.slave_open(self.ptyname)

        # Set the size of the pty
        # FIXME: There is a large chance that this is Linux-specific... :(
        winsize = struct.pack("HHHH", height, width, 0, 0)
        fcntl.ioctl(self.masterfd, IOCTLS.TIOCSWINSZ, winsize)

        # Fork off another process
        self.pid = os.fork()
        if self.pid < 0:
            # FIXME: fork failed, die here!
            pass
        if self.pid == 0:
            # we are the child process now
            
            # connect up to the pty
            os.dup2(self.slavefd, 0)
            os.dup2(self.slavefd, 1)
            os.dup2(self.slavefd, 2)
            
            # and fire off the program
            os.execv("/bin/sh", ["sh", "-c", "exec " + cmd])
            os._exit(-1)
            # this will never be reached
        # we are the parent process -- nothing to do
    def __del__(self):
        import os

        # kill the child process
        os.kill(self.pid, 15)
        
        # close the ptys
        os.close(self.slavefd)
        os.close(self.masterfd)

# Control key translation
keytranslations = { curses.key_left : chr(27) + "[D", \
                    curses.key_right : chr(27) + "[C", \
                    curses.key_up : chr(27) + "[A", \
                    curses.key_down : chr(27) + "[B", \
                    curses.key_backspace : chr(8), \
                    curses.key_ic : chr(27) + "[2~", \
                    curses.key_dl : chr(27) + "[3~", \
                    curses.key_home : chr(27) + "[1~", \
                    curses.key_end : chr(27) + "[4~", \
                    curses.key_ppage : chr(27) + "[5~", \
                    curses.key_npage : chr(27) + "[6~", \
                    curses.key_f1 : chr(27) + "[[A", \
                    curses.key_f2 : chr(27) + "[[B", \
                    curses.key_f3 : chr(27) + "[[C", \
                    curses.key_f4 : chr(27) + "[[D" }

# Wrapper class for stdin along the same lines as TerminalWindow
class Dispatcher:
    def __init__(self):
        self.state = 0
    def fileno(self):
        return sys.stdin.fileno()
    def process(self):
        self.dispatch(activeterminal.win.getch())
    def dispatch(self, key):
        if self.state == 0:
            try:
                str = chr(key)
            except ValueError:
                # Translate control keys to what our terminfo says
                if keytranslations.has_key(key):
                    activeterminal.feed(keytranslations[key])
                # else... just ignore it, I suppose!
            else:
                activeterminal.feed(str)

# Set up the root window
win = curses.initscr()
curses.cbreak()
curses.noecho()
curses.nonl()
curses.intrflush(0)
curses.meta(1)

# Create the status bar at the bottom
statwin = StatusWindow(win)

# Make sure that our children get the right value for TERM
os.environ['TERM'] = "superterm"

# Create the dispatcher
disp = Dispatcher()

# Create a single terminal for now
y, x = win.getmaxyx()
term = TerminalWindow(y-1, x, 0, 0, "/bin/bash")
activeterminal = term

# Handle events
while not quityet:
    try:
        a, b, c = select.select([disp, term], [], [], 1.0)
        for x in a:
            x.process()
        statwin.process()
        term.win.refresh()
    except KeyboardInterrupt:
	disp.dispatch(3)
        #quityet = 1

# Shut down curses
curses.endwin()

print "\n\n[superterm exiting]\n"




